#include "wepgen.h"
//
// constructors ...
//
/*----------------------------------------------------------------
* nwepgen
*
* Generates a set of WEP keys from a generator string.  This is 
* intended as a convenience.  Entering hex bytes can be a pain.
*
* Based on an algorithm supplied by Neesus Datacom, 
* http://www.neesus.com
*
* This function was authored by Zoom Telephonics Engineer 
* Juan Arango.
* http://www.zoomtel.com
*
* Juan's Note: 
* Changing the code in this function could make this product 
* incompatible with other ZoomAir wireless products because 
* these other products rely on Microsoft's rand() and srand() 
* function implementations!!!  This code uses the same algorithm.
*
* Distributed with permission from Zoom Telephonics.
*
* Arguments:
* 	genstr		a null terminated string
*	keylen		number of bytes in key
* 	wep_key		a 2d array that is filled with the wep keys
* Returns:
*	nothing
----------------------------------------------------------------*/
#define WLAN_WEP_NKEYS      4
#define WLAN_WEP_MAXKEYLEN 13
typedef unsigned char UINT8 ;
wepgen::wepgen() : Type( Wep_TypeUnknown ) {
    memset( (void*)&(Octets[0]), 0x00,sizeof(this->Octets) ) ;
}
wepgen::wepgen( const caddr_t& str, enum WepType type ) : Type(type) {

    memset( (void*)&(Octets[0]), 0x00,sizeof(this->Octets) ) ;
    
    switch (Type) {
        case Wep_1x64b  :
        case Wep_4x64b  : wep_keygen40( str )  ; break ;
        case Wep_1x128b :
        case Wep_4x128b : wep_keygen128( str ) ; break ;
        default         : break ;
    }

}
void wepgen::toStdout() const {
    for ( unsigned char i = 0 ; i < Type ; i++ ) {
        fprintf( stdout, "%02x", Octets[i] ) ;
    }
}
/*
 * generate four subkeys from a seed using the defacto standard
 */
void wepgen::wep_seedkeygen(int val) {

    for(int i = 0; i < WEPKEYSTORE; i++) {
        val *= 0x0343fd;
        val += 0x269ec3;
        Octets[i] = val >> 16;
    }

    return;
}
/*
 * generate one key from a string using the de facto standard
 *
 * resultant key is stored in
 *   one 128 bit key: keys[0-15]
 *
 * (NOTE: I'm not sure why, but it seems that only values 0-12 are used,
 * resulting in 104 bits of keying, not 128)
 */
void wepgen::wep_keygen128(const caddr_t& str) {

    u_char buf[64];

    //
    // This bit of turdage is to insure against a stack basher coming
    // in with redirect code in str, and to keep the gd compiler happy.
    //
    for (
          caddr_t s = str, p = (caddr_t)buf
        ; *s && ((unsigned int)(p - (caddr_t)buf) < (unsigned int)sizeof( buf ) )
        ; s = *s ? s : (caddr_t)str ) { 

        *p++ = *s++ ;

    }

    struct md5_ctx ctx;
    md5_init_ctx(&ctx) ;
    md5_process_block(buf,sizeof(buf),&ctx);
    md5_finish_ctx(&ctx, (void*)buf);

    int i = 0 ;
    for( ; i < WEPSTRONGKEYSIZE; i++) {
        Octets[i] = buf[i];
    }
    for( ; i < WEPKEYSTORE; i++) {
        Octets[i] = 0;
    }
    return;
}

/* 
 * generate four subkeys from a string using the defacto standard
 *
 * resultant keys are stored in 
 *   four 40 bit keys: keys[0-4], keys[5-9], keys[10-14] and keys[15-20]
 */
void wepgen::wep_keygen40(const caddr_t& str) {

    /*
     * seed is generated by xor'ing in the keystring bytes
     * into the four bytes of the seed, starting at the little end
     */

    int val = 0 ;
    for(int i = 0, shift = 0 ; str[i]; i++) {
        shift = i & 0x3;
        val ^= (str[i] << (shift * 8));
    }

    wep_seedkeygen(val);

    return;
}
#define STANDALONE
#ifdef STANDALONE
void Usage( const char* app ) {
    fprintf( stderr, "Usage: %s [4x64b|1x128b] wep_string\n", app ) ;
}
int main( int argc, char** argv ) {

    if (argc != 3) {
        Usage( *argv ) ;
        return -1 ;
    }

    unsigned int myKeyLength =
        ! strcmp( "1x128b", argv[1] ) ? wepgen::Wep_1x128b :
        ! strcmp(  "4x64b", argv[1] ) ? wepgen::Wep_4x64b  : 0 ;
        ;

    if ( ! myKeyLength ) {
        Usage( *argv ) ;
        return -1 ;
    }

    wepgen s(
         (const caddr_t)argv[2]
        ,myKeyLength == wepgen::Wep_1x128b ?
            wepgen::Wep_1x128b : wepgen::Wep_4x64b ) ;

    s.toStdout() ; 

    printf( "\n" ) ;

    return 0 ;

}
#endif
